;
; BFFCOMMS - Serial COMMS for PICAXE - functions. Version beta 02
;
; v beta 02 - corrected BFFCOMMS_Purge() arguments in text
;
; Ian Hopper - BFF Design Ltd - June 2010
; Ian@builtforfun.co.uk - http://builtforfun.co.uk/
;
;  2010, 2011 BFF Design Ltd
; 
;======================================================================================
; This file is part of BFF Serial Comms for PICAXE.
;
;    BFF Serial Comms for PICAXE is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 3 of the License, or
;    (at your option) any later version.
;
;    BFF Serial Comms for PICAXE is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
;======================================================================================
;
;The BFFCOMMS functions are :-
;
;Open the COM port
;=================
;
;BFFCOMMS_Open(COMPort, Baud, Handle, TimeOut, Abort_Disable)
;   COMPort - Input String, port number - eg COM1, COM2 etc
;   Baud - Input Integer, baud rate - eg 9600, 115200 etc
;   Handle - Output Integer, handle for the opened port
;   Timeout - Optional Input Integer, read timeout period in ms (defaults to zero - no timeout)
;   Abort_Disable - Optional Input Integer, =1 to disable Abort on Error, = 0 to enable
;
;   Return integer, = 0 for success, >= 0 gives windows error code for last error
;
;
;
;Close the COM port
;==================
;
;BFFCOMMS_Close(Handle)
;   Handle - Input Integer, handle for the opened port (from BFFCOMMS_Open() function call)
;
;   Return integer, = 0 for success, >= 0 gives windows error code for last error
;
;
;
;Purge the COM port read buffer
;==============================
;
;BFFCOMMS_Purge(Handle)
;   Handle - Input Integer, handle for the opened port
;
;   Return integer, = 0 for success, >= 0 gives windows error code for last error
;
;
;
;Send byte string up to 16 bytes long to the COM port
;====================================================
;
;BFFCOMMS_Send(Handle, dFormat, b1, b2, b3........ b16)
;   Handle - Input Integer, handle for the opened port
;   dFormat - Input String, of form AABBBB etc, length gives number of bytes, character value gives type - A=ascii, B=data, C=carriage return, L=line feed
;   b1 - Output byte, data byte value 0-255 or single ascii character
;   b2 to b16 - Optional output bytes
;
;   Return integer, = 0 for success, =-1 zero length format definition, >= 0 gives windows error code for last error
;
;
;
;Get byte string up to 16 bytes long from the COM port
;=====================================================
;
;BFFCOMMS_Get(Handle, dFormat, b1, b2, b3........ b16)
;   Handle - Input Integer, handle for the opened port
;   dFormat - Input String, of form AABBBB etc, length gives number of bytes, character value gives type - Q=ascii qualifier, A=ascii, B=data, C=carriage return, L=line feed
;     Q qualifiers when present should be the first characters in dFormat, 
;   b1 - Output byte, data byte value 0-255 or single ascii character
;   b2 to b16 - Optional output bytes
;
;   Return integer, = 0 for success, = -1 timeout has occured, = -2 zero format string length, >= 0 gives windows error code for last error
;
;NOTE if ascii quailifiers are found in the serial data stream they will be echoed in the relevant returned bytes.
;The timeout behaviour of the function when it is listening for qualifier bytes is determined by the global variable
;BFFCOMMS_QTO. If BFFCOMMS_QTO = 0 then the function will wait indefinately for the qualifier bytes to appear in the
;data stream. If BFFCOMMS_QTO = 1 then the function will timeout if no data is recieved within the timeout period set in
;function BFFCOMMS_Open() whilst it waits for qualifier bytes.
;
;If a timeout is specified in BFFCOMMS_Open() then it will always be active when the function is listening for non-qualifier
;bytes.
;
;IMPORTANT - if your serial cable is present but not connected to a PICAXE download circuit then the line voltages might float
;and the BFFCOMMS_Get() function may return "ghost" byte values under some circumstances.
;


;=========================================================================
;
;Setup the read/write COM port
;
;=========================================================================
BFFCOMMS_Open(COMPort, Baud, ByRef Handle, TimeOut = 0, Abort_Disable = 0)
{
  Global BFFCOMMS_TCNT

  FileAppend, BFFCOMMS_Open: Called %COMPort% %Baud% %Timeout% %Abort_Disable%`n, BFF_Comms.log

;Constants
  
  iret := 0
  GENERIC_WRITE = 0x40000000  	;Open the port for writing
  GENERIC_READ = 0x80000000 	;For Reading
  GENERIC_ALL = 0x10000000
  iOpen := 3	; File to be opened must already exist.

;Open the COM port

  Handle := DllCall("CreateFile", "Str", COMPort, "Uint", GENERIC_READ | GENERIC_WRITE, "Uint", 0, "UInt", 0, "UInt", iOpen, "Uint", 0, "UInt", 0)
  if (Handle < 1 OR ErrorLevel != 0)
  {
  	ErrorLevel1 := ErrorLevel
  	err := DllCall("GetLastError")
  	if (err = 2)
  	{
  	 MsgBox, Error on CreateFile(), Specified COM port %COMPort% not found......
  	 FileAppend, BFFCOMMS_Open: Error on CreateFile() Specified COM port %COMPort% not found......`n, BFF_Comms.log
  	 return, err
  	}
    else
  	{
  	 MsgBox, Error on CreateFile() Handle = %Handle%, ErrorLevel = %ErrorLevel1%, Win Error= %err%
  	 FileAppend, BFFCOMMS_Open: Error on CreateFile() Handle = %Handle%, ErrorLevel = %ErrorLevel1%, Win Error= %err%`n, BFF_Comms.log
  	 return, err
  	}
  }

  FileAppend, BFFCOMMS_Open: Port opened cleanly...%Handle%`n, BFF_Comms.log

;
;Now setup the com port state
;
;Read the current state to setup the DCB port config structure - the relevant data will be adjusted below then re-written

  n_995_err := 0

Try_GCS_again:
  
  err := 0
  n_995_err += 1

  VarSetCapacity(DCB, 28, 0)    ;reserve space for the DCB structure data

  iret1 := DllCall("GetCommState", "Uint", Handle, "Str", DCB)
  if (iret1 = 0 OR ErrorLevel != 0)
  {
  	ErrorLevel1 := ErrorLevel
  	err := DllCall("GetLastError")
	
  	if (err = 995)   ;if this is an abort error then try again up to five times to get the com state
  	 {
      if (n_995_err < 6)
      {
        FileAppend, BFFCOMMS_Open: Error on GetCommState() iret = %iret1% ErrorLevel = %ErrorLevel1% Win Error= %err%...trying again`n, BFF_Comms.log
        DllCall("Sleep", UInt, 100)
        goto Try_GCS_again
      }
     }

  	MsgBox, Error on GetCommState() iret = %iret1%, ErrorLevel = %ErrorLevel1%, Win Error= %err%
  	FileAppend, BFFCOMMS_Open: Error on GetCommState() iret = %iret1% ErrorLevel = %ErrorLevel1% Win Error= %err% n_995_err=%n_995_err% ...leaving`n, BFF_Comms.log
  	return, 1
  }

  FileAppend, BFFCOMMS_Open: Comm State read...`n, BFF_Comms.log

;
;Update the COM port state to set the 8N1 format by modifying the DCB structure
;

;Set port baud
  NumPut(Baud, DCB, 4, "Int")

;Set flow control to off
  if (Abort_Disable = 1)
    NumPut(4241, DCB, 8, "Int")      ; 4214 = 0001 0000 1001 0001      ;with fAbortOnError set to zero to remove 995 errors
  else
    NumPut(20625, DCB, 8, "Int")     ; 20625 = 0101 0000 1001 0001      ;with fAbortOnError set to allow 995 errors

;Set Byte Size to 8, Parity to None & Stop Bits to 1
  NumPut(8, DCB, 18, "Uchar")
  NumPut(0, DCB, 19, "Uchar")
  NumPut(0, DCB, 20, "Uchar")

;Call SetCommState to set the new port state with the modified DCB

  iret2 := DllCall("SetCommState", "Uint", Handle, "Str", DCB)
  if (iret2 = 0 OR ErrorLevel != 0)
  {
  	ErrorLevel1 := ErrorLevel
  	err := DllCall("GetLastError")
  	MsgBox, Error on SetCommState() iret = %iret2%, ErrorLevel = %ErrorLevel1%, Win Error= %err%
  	FileAppend, BFFCOMMS_Open: Error on SetCommState() iret = %iret2% ErrorLevel = %ErrorLevel1% Win Error= %err%`n, BFF_Comms.log
  	return, err
  }

  FileAppend, BFFCOMMS_Open: Comm State written...`n, BFF_Comms.log

;
;Now set the read timeout
;

;Read the structure
  VarSetCapacity(CommTimeouts, 24, 0)

  iret3 := DllCall("GetCommTimeouts", "UInt", Handle, "Str", CommTimeouts)
  if (iret3 = 0)
    {
      err := DllCall("GetLastError")
      MSgBox GetCommTimeouts() error %iret3% Win error %err%
      FileAppend, BFFCOMMS_Open: GetCommTimeouts() error %iret3% Win error %err%`n, BFF_Comms.log
      return, err
    }

;Set values
  NumPut(0, CommTimeouts, 0, "UInt")
  NumPut(0, CommTimeouts, 4, "UInt")
  NumPut(Timeout, CommTimeouts, 8, "UInt")    ;read timeout in ms, 0 indicates no read timeouts used for the Port
  NumPut(0, CommTimeouts, 12, "UInt")
  NumPut(0, CommTimeouts, 16, "UInt")

;and re-write
  iret4 := DllCall("SetCommTimeouts", "UInt", Handle, "Str", CommTimeouts)
  if (iret4 = 0)
    {
      err := DllCall("GetLastError")
      MSgBox SetCommTimeouts() error %iret4% Win error %err%
      FileAppend, BFFCOMMS_Open: SetCommTimeouts() error %iret4% Win error %err%`n, BFF_Comms.log
      return, err
    }

  FileAppend, BFFCOMMS_Open: SetCommTimeouts() set...`n, BFF_Comms.log
  FileAppend, BFFCOMMS_Open: Exit cleanly...`n, BFF_Comms.log
  
;Set the global timeout tick counter setting

  iret5 := DllCall("QueryPerformanceFrequency", "Int64*", Tick_Cal)
  BFFCOMMS_TCNT := Round(Timeout / 1000.0 * Tick_Cal)

;That's it - return

return, 0
}



;=========================================================================
;
;Close the COM port
;
;=========================================================================
BFFCOMMS_Close(Handle)
{
  iret := DllCall("CloseHandle", "UInt", Handle)
  
  ;Check Error

  if (iret = 0)
  {
    err := DllCall("GetLastError")
    FileAppend, BFFCOMMS_Close: Error on CloseHandle() iret = %iret%, Win Error= %err%`n, BFF_Comms.log
    return, err
  }

  FileAppend, BFFCOMMS_Close: COM Port closed cleanly...`n, BFF_Comms.log
  
return, 0
}


  
  
;=========================================================================  
;
;Send up to 16 bytes to the opened COM port with handle Handle. At least 1 byte must be sent
;
;Format gives the byte string format - it is a string n_bytes long, each character needs to be A (ascii), B (binary), C (CR) OR L (LF) to indicate format of each byte
;
;=========================================================================
BFFCOMMS_Send(Handle, dFormat, b1, b2 = 0, b3 = 0, b4 = 0, b5 = 0, b6 = 0, b7 = 0
, b8 = 0, b9 = 0, b10 = 0, b11 = 0, b12 = 0, b13 = 0, b14 = 0, b15 = 0, b16 = 0)
{

  VarSetCapacity(buffer, 16, 0)  
  
;Get length of data

  n_bytes := Strlen(dFormat)
  if (n_bytes = 0)
    return, -1
  else if (n_bytes > 16)
    n_bytes = 16
  
;Parse the data

  Loop, %n_bytes%
  {
    byte_type := SubStr(dFormat, A_Index, 1)
    if (byte_type = "A")            ;send byte as ascii character
      {
      byte_ascii := Asc(b%A_Index%)       ;if byte is longer than 1 char then 1st character is used
      NumPut(byte_ascii, buffer, A_Index - 1, "UChar")
      }
    else if (byte_type = "C")      ;send CR
      NumPut(13, buffer, A_Index - 1, "UChar")
    else if (byte_type = "L")      ;send NL
      NumPut(10, buffer, A_Index - 1, "UChar")
    else                            ;"B" or unknown - send as binary value 0-255
      {
      if (b%A_Index% > 255)
        b%A_Index% := 255
      else if (b%A_Index% < 0)
        b%A_Index% := 0
      NumPut(b%A_Index%, buffer, A_Index - 1, "UChar")
      }
  }   
  
;Write the buffer to Handle

  iret := DllCall("WriteFile", "UInt", Handle, "UInt", &buffer, "UInt", n_bytes, "UIntP", LWritten, "UInt", 0)   
  if (iret = 0)
  {
  	ErrorLevel1 := ErrorLevel
  	err := DllCall("GetLastError")
  	MsgBox, Error on BFFCOMMS_Send() iret = %iret%, ErrorLevel = %ErrorLevel1%, Win Error= %err%
  	FileAppend, BFFCOMMS_Send: Error on SetCommState() iret = %iret% ErrorLevel = %ErrorLevel1% Win Error= %err%`n, BFF_Comms.log
  	return, err
  }  
  
;Return

return, 0
  
}  
  
  
;=========================================================================  
; 
;Get up to 16 bytes from the opened COM port with handle Handle  
;
;=========================================================================
BFFCOMMS_Get(Handle, dFormat, ByRef b1, ByRef b2 = 0, ByRef b3 = 0, ByRef b4 = 0, ByRef b5 = 0, ByRef b6 = 0, ByRef b7 = 0
, ByRef b8 = 0, ByRef b9 = 0, ByRef b10 = 0, ByRef b11 = 0, ByRef b12 = 0, ByRef b13 = 0, ByRef b14 = 0, ByRef b15 = 0, ByRef b16 = 0)
{  
  Global BFFCOMMS_QTO
  
;size the buffer
   
  VarSetCapacity(buffer, 32, 0)  
  
;Get length of data

  n_bytes := Strlen(dFormat)
  if (n_bytes = 0)
    return, -2
  else if (n_bytes > 16)
    n_bytes = 16  
    
;Any qualifiers in the format definition? Search from the start - 1st non qualifier is taken as end of qualifiers even if more appear later

  n_qual := 0
  Loop, %n_bytes%
  {
    byte_type := SubStr(dFormat, A_Index, 1)
    if (byte_type = "Q")
      {
      n_qual += 1
      qual%n_qual% := b%A_Index%
      }
    else
      break
  }
  n_data := n_bytes - n_qual      ;the rest must be data bytes

;Look for the qualifiers in the serial data,

  if (n_qual > 0)
  {
    k_qual := 1
    While (k_qual <= n_qual)
    {
      iret2 := BFFCOMMS_Getbyte(Handle, byte)
      if (iret2 = -1 AND BFFCOMMS_QTO = 1)
        return, -1
      else if (iret2 != 0 AND iret2 != -1)
        return, iret2
      byte_char := Chr(byte)
      if (byte_char = qual%k_qual%)   ;found one
        k_qual += 1   ;go look for next
      else
        k_qual := 1   ;start again
    }
  }

;Now get the data following the qualifiers (if there were any), always come out if times-out

  k_data := n_qual + 1
  Loop, %n_data%
  {
    iret2 := BFFCOMMS_Getbyte(Handle, byte)
    if (iret2 = -1)
      return, -1
    else if (iret2 != 0)
      return, iret2
    byte_type := SubStr(dFormat, k_data, 1)
    if (byte_type = "A")
      b%k_data% := Chr(byte)
    else
      b%k_data% := byte
    
    k_data += 1   
  }

;Return

return, 0
  
}  



;=========================================================================
;
;Get one byte from the com port
;
;=========================================================================
BFFCOMMS_Getbyte(Handle, ByRef byte)
{
  Global BFFCOMMS_TCNT
  
;Read one byte from handle

  VarSetCapacity(Data, 1, 0) 
  DllCall("QueryPerformanceCounter", "Int64*", T_Before) 
  iret := DllCall("ReadFile", "UInt", Handle, "UInt", &Data, "UInt", 1, "UIntP", LRead, "UInt", 0)
  DllCall("QueryPerformanceCounter", "Int64*", T_After)

;Check for read Error

  if (iret = 0)
  {
    err := DllCall("GetLastError")
    FileAppend, BFFCOMMS_Getbyte: Error on ReadFile() iret = %iret%, Win Error= %err%`n, BFF_Comms.log
    return, err
  }

;Check the time taken for the read - ie has the read timed-out?

  T_Read := T_After - T_Before
  if (BFFCOMMS_TCNT != 0 AND T_Read >= 0.95 * BFFCOMMS_TCNT)
    Return, -1

;Set the byte
  
  byte := NumGet(Data, 0, "UChar")

return, 0
}



;=========================================================================
;
;Purge the read buffer
;
;=========================================================================
BFFCOMMS_Purge(Handle)
{

;Purge

  iret := DllCall("PurgeComm", "UInt", Handle, "UInt", 0x0008)

;Check Error

  if (iret = 0)
  {
    err := DllCall("GetLastError")
    FileAppend, BFFCOMMS_Purge: Error on PurgeComm() iret = %iret% Win Error= %err%`n, BFF_Comms.log
    return, err
  }
 
return, 0
}


    